Imports System.Math

' An object that represents a expression parse tree.
Public Class CompiledExpression
    ' A dictionary of variable name/value pairs.
    Public Variables As New Dictionary(Of String, Double)

    ' Operators.
    Private Enum OperatorTypes
        op_Constant ' A numeric constant.
        op_Variable ' A variable.
        op_Parens   ' Parentheses as in (17).
        op_Neg      ' Unary minus as in 3 * -2.
        op_Pos      ' Unary plus as in 2 * +3.
        op_Power    ' Power as in 3^2.
        op_Times    ' Times as in 6 * 7.
        op_Div      ' Division as in 4 / 5.
        op_IntDiv   ' Integer division as in 4 \ 5.
        op_Mod      ' Mod as in 7 % 4.
        op_Plus     ' Binary plus as in 3 + 9.
        op_Minus    ' Binary minus as in 3 - 9.
        op_Sin      ' Sine.
        op_Cos      ' Cosine.
        op_Tan      ' Tangent.
        op_Sqrt     ' Square root.
        op_Log10    ' Log base 10.
        op_Log      ' Natural log.
        op_Exp      ' Exponential
        op_ATan     ' ArcTan.
        op_ASin     ' ArcSin.
        op_ACos     ' ArcCos.
    End Enum

    ' Operator precedence.
    Private Enum PrecedenceTypes
        prec_Plus = 1
        prec_Mod = 2
        prec_IntDiv = 3
        prec_Times = 4
        prec_Unary = 5
        prec_Power = 6
        prec_None = 7
    End Enum

    Private LeftOperand As CompiledExpression
    Private RightOperand As CompiledExpression
    Private OperatorType As OperatorTypes
    Private ConstantValue As Double
    Private VariableName As String

    ' Parse the expression text and compile it.
    Public Sub New(ByVal expr As String)
        ' Split the expression and see what kind of expression it is.
        Dim left_operand As String = Nothing
        Dim right_operand As String = Nothing

        SplitExpression(expr, OperatorType, left_operand, right_operand)

        ' See what kind of operation we need to apply.
        If OperatorType = OperatorTypes.op_Constant Then
            ' It's a constant. Evaluate it now.
            ConstantValue = Double.Parse(left_operand)
        ElseIf OperatorType = OperatorTypes.op_Variable Then
            ' It's a variable. Save its name.
            VariableName = left_operand
        Else
            ' Some other "real" operation. 
            ' Make the left operand.
            LeftOperand = New CompiledExpression(left_operand)

            ' See if we need a right operand.
            If Not (right_operand Is Nothing) Then
                RightOperand = New CompiledExpression(right_operand)
            End If
        End If
    End Sub

    ' Evaluate the parsed expression tree.
    Public Function Evaluate() As Double
        Return Me.Evaluate(Variables)
    End Function
    Public Function Evaluate(ByVal variable_values As Dictionary(Of String, Double)) As Double
        Select Case OperatorType
            Case OperatorTypes.op_Constant ' A numeric constant.
                Return ConstantValue
            Case OperatorTypes.op_Variable ' A variable.
                Return variable_values(VariableName)
            Case OperatorTypes.op_Parens   ' Parentheses as in (17).
                Return LeftOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Neg      ' Unary minus as in 3 * -2.
                Return -LeftOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Pos      ' Unary plus as in 2 * +3.
                Return LeftOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Power    ' Power as in 3^2.
                Return LeftOperand.Evaluate(variable_values) ^ RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Times    ' Times as in 6 * 7.
                Return LeftOperand.Evaluate(variable_values) * RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Div      ' Division as in 4 / 5.
                Return LeftOperand.Evaluate(variable_values) / RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_IntDiv   ' Integer division as in 4 \ 5.
                Return CLng(LeftOperand.Evaluate(variable_values)) \ CLng(RightOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Mod      ' Mod as in 7 % 4.
                Return LeftOperand.Evaluate(variable_values) Mod RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Plus     ' Binary plus as in 3 + 9.
                Return LeftOperand.Evaluate(variable_values) + RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Minus    ' Binary minus as in 3 - 9.
                Return LeftOperand.Evaluate(variable_values) - RightOperand.Evaluate(variable_values)
            Case OperatorTypes.op_Sin      ' Sine.
                Return Sin(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Cos      ' Cosine.
                Return Cos(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Tan      ' Tangent.
                Return Tan(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Sqrt     ' Square root.
                Return Sqrt(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Log10    ' Log 10.
                Return Log10(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Log      ' Log.
                Return Log(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_Exp      ' Exponential.
                Return Exp(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_ATan     ' ArcTan.
                Return Atan(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_ASin     ' ArcSin.
                Return Asin(LeftOperand.Evaluate(variable_values))
            Case OperatorTypes.op_ACos     ' ArcCos.
                Return Acos(LeftOperand.Evaluate(variable_values))
        End Select
    End Function

    ' Split the expression into an operator and operands.
    Private Sub SplitExpression(ByVal expr As String, ByRef op As OperatorTypes, ByRef left_operand As String, ByRef right_operand As String)
        Const digits As String = "0123456789."
        Const operator_chs As String = "^*/\%+-"
        Dim operators() As OperatorTypes = { _
            OperatorTypes.op_Power, _
            OperatorTypes.op_Times, _
            OperatorTypes.op_Div, _
            OperatorTypes.op_IntDiv, _
            OperatorTypes.op_Mod, _
            OperatorTypes.op_Plus, _
            OperatorTypes.op_Minus}
        Dim open_parens As Long
        Dim ch As String
        Dim is_unary As Boolean     ' If ch is +/-, it is unary.
        Dim next_unary As Boolean   ' If the next ch is +/-, it is unary.
        Dim new_prec As PrecedenceTypes
        Dim best_prec As PrecedenceTypes
        Dim best_ch As String = ""
        Dim best_i As Integer
        Dim function_name As String
        Dim test_value As Double

        ' Remove leading and trailing whitespace and internal spaces.
        expr = expr.Trim().Replace(" ", "")
        Dim expr_len As Integer = Len(expr)
        If expr_len = 0 Then Throw New ArgumentNullException("expr")

        ' If the first character is +/-, it is unary.
        next_unary = True
        best_prec = PrecedenceTypes.prec_None

        ' Examine each character to find the operator
        ' with the lowest precedence. This is the
        ' operator that we evaluate last.
        For i As Integer = 0 To expr_len - 1
            ' Get ready to examine the next character.
            is_unary = next_unary

            ' Examine the next character.
            ch = expr.Substring(i, 1)
            If ch = "(" Then        ' Increment open paren count.
                open_parens += 1
                next_unary = True       ' +/- after ( is unary.
            ElseIf ch = ")" Then        ' Decrement open paren count.
                open_parens -= 1
                If open_parens < 0 Then Throw New ParenthesesException("Too many close parentheses")
                next_unary = False      ' +/- after ) is binary.
            ElseIf open_parens = 0 Then ' Could be an operator here.
                ' See if this is an operator.
                If operator_chs.IndexOf(ch) >= 0 Then
                    ' An operator after an operator is unary.
                    next_unary = True

                    ' See what precedence this operator has.
                    Select Case ch
                        Case "^"
                            new_prec = PrecedenceTypes.prec_Power
                        Case "*", "/"
                            new_prec = PrecedenceTypes.prec_Times
                        Case "\"
                            new_prec = PrecedenceTypes.prec_IntDiv
                        Case "%"
                            new_prec = PrecedenceTypes.prec_Mod
                        Case "+", "-"
                            If is_unary Then
                                If i = 0 Then
                                    ' The expression starts with a unary operator.
                                    new_prec = PrecedenceTypes.prec_Unary
                                Else
                                    ' Ignore internal unary operators.
                                    new_prec = PrecedenceTypes.prec_None
                                End If
                            Else
                                new_prec = PrecedenceTypes.prec_Plus
                            End If
                    End Select

                    ' See if this beats the previous best.
                    If new_prec <= best_prec Then
                        best_prec = new_prec
                        best_ch = ch
                        best_i = i
                    End If
                Else
                    ' This is a non-operator such as X or 3.
                    ' A following +/- is binary.
                    next_unary = False
                End If
            End If ' End examining ch.
        Next i ' End examining all characters.

        ' Make sure the parentheses count is zero.
        If open_parens <> 0 Then Throw New ParenthesesException("Mismatched parentheses")

        ' See what we've learned.
        If best_prec = PrecedenceTypes.prec_Unary Then
            ' A unary operator. See which one.
            If expr.StartsWith("-") Then
                ' Expr has form -expr2.
                left_operand = expr.Substring(1)
                op = OperatorTypes.op_Neg
            Else
                ' Expr has form +expr2.
                left_operand = expr.Substring(1)
                op = OperatorTypes.op_Pos
            End If
        ElseIf best_prec < PrecedenceTypes.prec_None Then
            ' We found some other operator.
            left_operand = expr.Substring(0, best_i)
            right_operand = expr.Substring(best_i + 1)
            op = operators(operator_chs.IndexOf(best_ch))
        ElseIf expr.StartsWith("(") And expr.EndsWith(")") Then
            ' Expr has form (expr2).
            left_operand = expr.Substring(1, expr.Length - 2)
            op = OperatorTypes.op_Parens
        ElseIf expr.EndsWith(")") Then
            ' Expr has form Func(expr2). 
            ' See which function.
            function_name = expr.Split("("c)(0).ToLower

            Select Case function_name
                Case "sin"
                    op = OperatorTypes.op_Sin
                Case "cos"
                    op = OperatorTypes.op_Cos
                Case "tan"
                    op = OperatorTypes.op_Tan
                Case "sqrt"
                    op = OperatorTypes.op_Sqrt
                Case "log10"
                    op = OperatorTypes.op_Log10
                Case "log"
                    op = OperatorTypes.op_Log
                Case "exp"
                    op = OperatorTypes.op_Exp
                Case "atan"
                    op = OperatorTypes.op_ATan
                Case "acos"
                    op = OperatorTypes.op_ACos
                Case "asin"
                    op = OperatorTypes.op_ASin
                Case Else
                    ' Unknown function (or bad syntax).
                    Throw New ArgumentException( _
                        "Cannot parse '" & expr & "'")
            End Select
            left_operand = expr.Substring(function_name.Length + 1, expr.Length - function_name.Length - 2)
        ElseIf digits.IndexOf(expr.Substring(0, 1)) >= 0 Then
            ' Expr is a numeric constant.
            ' Validate the value.
            Try
                test_value = Double.Parse(expr)
            Catch exc As Exception
                Throw New ArgumentException( _
                    "Cannot parse '" & expr & "' as a constant")
            End Try

            left_operand = expr
            op = OperatorTypes.op_Constant
        Else
            ' It must be a variable.
            left_operand = expr
            op = OperatorTypes.op_Variable
        End If
    End Sub ' SplitExpression

    ' Thrown when there is a mismatch in parentheses.
    Public Class ParenthesesException
        Inherits Exception

        Public Sub New()
            Me.New("Mismatched paremtheses")
        End Sub

        Public Sub New(ByVal message As String)
            MyBase.New(message)
        End Sub

        Public Sub New(ByVal message As String, ByVal inner As Exception)
            MyBase.New(message, inner)
        End Sub
    End Class ' ParenthesesException
End Class
